# Module for Centerline creation for vessels

import numpy as np

from loguru import logger
from kimimaro import skeletonize
from scipy.interpolate import interp1d
from scipy.signal import savgol_filter

from mircato_stats.statistics.cpr import _compute_tangent_vectors

# These are the allowed arguments for kimimaro.skeletonize()
# These are specifically for the teasar_params dictionary argument
TEASAR_KWARGS = {
    "scale",
    "const",
    "pdrf_scale",
    "pdrf_exponent",
    "soma_acceptance_threshold",
    "soma_detection_threshold",
    "soma_invalidation_const",
    "soma_invalidation_scale",
    "max_paths",
}
# These are the rest of the arguments
NON_TEASAR_KWARGS = {
    "dust_threshold",
    "progress",
    "fix_branching",
    "in_place",
    "fix_borders",
    "parallel",
    "parallel_chunk_size",
    "extra_targets_before" "extra_targets_after",
    "fill_holes",
    "fix_avocados",
    "voxel_graph",
}
SKELETONIZE_KWARGS = TEASAR_KWARGS.union(NON_TEASAR_KWARGS)
SAMPLING_KWARGS = {"is_thoracic", "number_of_points", "window_length"}


def create_centerline(
    vessel: np.ndarray, voxel_spacing: tuple, vessel_label: int = 1, **kwargs
) -> np.ndarray | None:
    """Create a centerline for a given vessel using skeletonize_vessel and postprocess_skeleton
    :param vessel: vessel array
    :param voxel_spacing: voxel spacing
    :param vessel_label: vessel label
    :param kwargs: kwargs for skeletonize_vessel and postprocess_skeleton
    """
    skeleton_kwargs, sampling_kwargs = _validate_centerline_kwargs(kwargs)
    vertices, edges = skeletonize_vessel(
        vessel, voxel_spacing, vessel_label, **skeleton_kwargs
    )
    if vertices is None:
        return None
    centerline = postprocess_skeleton(vertices, edges, **sampling_kwargs)
    return centerline


def skeletonize_vessel(
    vessel: np.ndarray, voxel_spacing: tuple, vessel_label: int = 1, **kwargs
) -> tuple[np.ndarray, np.ndarray]:
    """Calculate the raw skeleton of the vessel. Returns skeleton vertices and edges
    :param vessel: the vessel numpy array
    :param voxel_spacing: the spacing of voxels in the array
    :param vessel_label: the label of the vessel in the array, default=1 (binary image)
    :param kwargs: additional keyword arguments for kimimaro.skeletonize() function
    :return: a tuple containing the centerline vertices and edges
    """
    # These are the 2 necessary parameters for teasar algorithm
    teasar_params = {"const": 40, "scale": 1.5}
    if kwargs:
        _validate_skeletonize_kwargs(kwargs)
        teasar_kwargs = _extract_teasar_kwargs(kwargs)
        teasar_params.update(teasar_kwargs)  # Update or add more teasar params
    # for each parameter, get the kwarg value, if it doesn't exist, use the default from the function
    skeleton = skeletonize(
        all_labels=vessel,
        teasar_params=teasar_params,
        anisotropy=voxel_spacing,
        object_ids=[vessel_label],
        dust_threshold=kwargs.get("dust_threshold", 1000),
        progress=kwargs.get("progress", False),
        fix_branching=kwargs.get("fix_branching", True),
        in_place=kwargs.get("in_place", False),
        fix_borders=kwargs.get("fix_borders", True),
        parallel=kwargs.get("parallel", 1),
        parallel_chunk_size=kwargs.get("parallel_chunk_size", 100),
        extra_targets_before=kwargs.get("extra_targets_before", []),
        extra_targets_after=kwargs.get("extra_targets_after", []),
        fill_holes=kwargs.get("fill_holes", False),
        fix_avocados=kwargs.get("fix_avocados", False),
        voxel_graph=kwargs.get("voxel_graph", None),
    )
    try:
        skel = skeleton[vessel_label]
        vertices = (
            skel.vertices / voxel_spacing
        )  # Divide by voxel spacing to have the units match
        edges = skel.edges
        return vertices, edges
    except KeyError:
        return None, None


def postprocess_skeleton(
    vertices: np.ndarray, edges: np.ndarray, **kwargs
) -> np.ndarray:
    """Postprocess the kimimaro skeleton into a final centerline
    :param vertices: the vertices of the centerline
    :param edges: the edges listing the connections between vertices
    :param is_thoracic: is the skeleton of the thoracic aorta
    :param kwargs: Additional optional arguments. Must be in [is_thoracic, number_of_points, window_length]
    :return: the final centerline as an (N, 3) numpy array
    """
    skeleton_length = vertices.shape[0]
    _validate_sampling_kwargs(kwargs)
    num_points, window_length = _get_sampling_values(skeleton_length, kwargs)
    ordered_centerline = _order_vertices(vertices, edges)
    spaced_centerline = _arclength_space_centerline(ordered_centerline)
    even_sampled_centerline = _evenly_sample_centerline(spaced_centerline, num_points)
    smoothed_centerline = _smooth_centerline(even_sampled_centerline, window_length)
    try:
        _compute_tangent_vectors(smoothed_centerline)
    except ValueError:
        # If computing tangent vectors fails, recursively reduce the number of centerline points until it works
        new_num_points = int(np.floor(num_points * 0.9))
        new_window_length = window_length - 1
        if new_window_length > new_num_points:
            new_window_length = new_num_points - 1
        return postprocess_skeleton(
            vertices,
            edges,
            number_of_points=new_num_points,
            window_length=new_window_length,
        )
    return smoothed_centerline


def _order_vertices(vertices: np.ndarray, edges: np.ndarray) -> np.ndarray:
    """Order the vertices of the skeleton from end to end by using the edges
    :param vertices: the vertices of the centerline
    :param edges: the edges listing the connections between vertices
    :return: the ordered centerline in the same shape as the skeleton
    """
    # Step 1: Build the adjacency list
    adjacency_list = {}
    for edge in edges:
        if edge[0] not in adjacency_list:
            adjacency_list[edge[0]] = []
        if edge[1] not in adjacency_list:
            adjacency_list[edge[1]] = []
        adjacency_list[edge[0]].append(edge[1])
        adjacency_list[edge[1]].append(edge[0])
    # Step 2: Identify the start (and end) vertex as it will only have 1 edge
    start_vertex = None
    for vertex, connected in adjacency_list.items():
        if len(connected) == 1:
            start_vertex = vertex
            break
    # Sanity check
    if start_vertex is None:
        raise ValueError("A start vertex could not be found.")
    # Step 3: Traverse the graph from the start vertex
    ordered_vertices_indices = [start_vertex]
    current_vertex = start_vertex
    # Since we know the length of the path, we can loop N times
    for _ in range(len(vertices) - 1):
        # The next vertex will be the one that is not the previous
        for vertex in adjacency_list[current_vertex]:
            if vertex not in ordered_vertices_indices:
                ordered_vertices_indices.append(vertex)
                break
        current_vertex = ordered_vertices_indices[-1]
    # Step 4: Map the ordered indices to the original vertices
    ordered_vertices = [vertices[idx] for idx in ordered_vertices_indices]
    return np.asarray(ordered_vertices)


def _arclength_space_centerline(ordered_centerline: np.ndarray) -> np.ndarray:
    """evenly space the centerline in array space
    :param ordered_centerline: the output of _order_vertices
    :return: the evenly spaced centerline
    """
    differences = np.diff(ordered_centerline, axis=0)
    distances = np.linalg.norm(differences, axis=1)
    arc_lengths = np.concatenate(([0], np.cumsum(distances)))
    total_length = arc_lengths[-1]
    normalized_arc_lengths = arc_lengths / total_length
    # Create interpolation functions for each coordinate
    interp_funcs = [
        interp1d(normalized_arc_lengths, ordered_centerline[:, i]) for i in range(3)
    ]
    # Create a new uniform parameterization
    uniform_t = np.linspace(0, 1, len(ordered_centerline))
    # Interpolate the centerline points
    uniform_centerline = np.column_stack([f(uniform_t) for f in interp_funcs])
    return uniform_centerline


def _evenly_sample_centerline(
    spaced_centerline: np.ndarray, number_of_points: int
) -> np.ndarray:
    """Evenly sample a specified number of points along the centerline
    :param spaced_centerline: the output of _arclength_space_centerline
    :param number_of_points: the number of points to sample from the centerline
    :return: the evenly sampled centerline of shape (number_of_points, 3)
    """
    # Calculate the cumulative distance along the centerline
    distances = np.cumsum(
        np.r_[0, np.linalg.norm(np.diff(spaced_centerline, axis=0), axis=1)]
    )
    total_length = distances[-1]

    # Create an interpolation function for each dimension
    f_z = interp1d(distances, spaced_centerline[:, 0])
    f_y = interp1d(distances, spaced_centerline[:, 1])
    f_x = interp1d(distances, spaced_centerline[:, 2])

    # Evenly sample distances along the centerline
    sample_points = np.linspace(0, total_length, number_of_points)
    # Sample the centerline
    sampled_centerline = np.column_stack(
        (f_z(sample_points), f_y(sample_points), f_x(sample_points))
    )
    return sampled_centerline


def _smooth_centerline(
    sampled_centerline: np.ndarray, window_length: int
) -> np.ndarray:
    """Use a window length to smooth the centerline using a savitzky-golay filter
    :param sampled_centerline: the sampled point centerline from _evenly_sample_centerline
    :param window_length: the length of the filter window
    :return: the smoothed centerline
    """
    n_dimensions = sampled_centerline.shape[1]
    polyorder = 2

    def _change_if_odd(window):
        if not window % 2:
            window += 1
        return window

    window_length = _change_if_odd(window_length)
    if window_length > len(sampled_centerline):
        window_length = _change_if_odd(int(len(sampled_centerline) // 2))
    if window_length < polyorder:
        logger.debug(
            f"Window length {window_length} is smaller than polyorder {polyorder}"
        )
        window_length = polyorder + 1
        # raise ValueError(f'Window length {window_length} is smaller than polyorder {polyorder}')
    smoothed_centerline = np.zeros_like(sampled_centerline)
    for dim in range(n_dimensions):
        smoothed_centerline[:, dim] = savgol_filter(
            sampled_centerline[:, dim], window_length, polyorder
        )
    smoothed_centerline = smoothed_centerline.round().astype(
        np.uint16
    )  # Make sure the centerline is integers
    return smoothed_centerline


def _get_sampling_values(skeleton_length: int, kwargs: dict) -> tuple[int, int]:
    """Get the number of points and window length for the sampling function using the original shape and kwargs
    :param skeleton_length: the original number of points in the skeleton, usually given as skeleton.shape[0]
    :param kwargs: a dictionary of keyword arguments with parameters necessary
    :return: a tuple containing the (number_of_points, window_length)
    """
    is_thoracic = kwargs.get("is_thoracic", False)
    has_specified_values = (
        kwargs.get("number_of_points") is not None
        and kwargs.get("window_length") is not None
    )
    if has_specified_values:
        number_of_points = kwargs.get("number_of_points")
        window_length = kwargs.get("window_length")
    elif is_thoracic:
        # For the thoracic aorta, we want fewer number of points and slightly wider window range
        number_of_points = int(skeleton_length // 2)
        window_length = int(number_of_points // 8)
    else:
        number_of_points = int(skeleton_length // 1.5)
        window_length = int(number_of_points // 10)
    return number_of_points, window_length


def _validate_centerline_kwargs(kwargs) -> tuple[dict, dict]:
    """Validate create_centerline keyword arguments and return the function specific arguments"""
    all_kwargs = SKELETONIZE_KWARGS.union(SAMPLING_KWARGS)
    assert kwargs == {} or all(k in all_kwargs for k in kwargs), ValueError(
        f"Invalid kwargs given to create_centerline: {kwargs}. Must be in {sorted(all_kwargs)}."
    )
    skeleton_kwargs = {k: v for k, v in kwargs.items() if k in SKELETONIZE_KWARGS}
    sampling_kwargs = {k: v for k, v in kwargs.items() if k in SAMPLING_KWARGS}
    return skeleton_kwargs, sampling_kwargs


def _validate_sampling_kwargs(kwargs):
    """Validate sampling keyword arguments"""
    assert kwargs == {} or all([k in SAMPLING_KWARGS for k in kwargs]), ValueError(
        f"Sampling kwargs must be in {SAMPLING_KWARGS}"
    )
    is_thoracic = kwargs.get("is_thoracic")
    number_of_points = kwargs.get("number_of_points")
    window_length = kwargs.get("window_length")
    # Check thoracic first
    if is_thoracic is not None:
        assert is_thoracic in [True, False], ValueError(
            f'is_thoracic must be a boolean: {kwargs.get("is_thoracic")=}'
        )
    if is_thoracic is None:
        assert number_of_points is not None and window_length is not None, ValueError(
            "number_of_points and window_length must be given if is_thoracic is None"
        )
    elif number_of_points is not None:
        assert isinstance(number_of_points, int), ValueError(
            "number_of_points must be an integer"
        )
        assert isinstance(window_length, int), ValueError(
            "window_length must be an integer provided with number_of_points"
        )
    elif window_length is not None:
        assert isinstance(window_length, int), ValueError(
            "window_length must be an integer"
        )
        assert isinstance(number_of_points, int), ValueError(
            "number_of_points must be an integer provided with window_length"
        )


def _validate_skeletonize_kwargs(kwargs: dict) -> None:
    """Validate skeletonize keyword arguments"""
    assert all([k in SKELETONIZE_KWARGS for k in kwargs]), ValueError(
        f"All kwargs for create_centerline must be in {SKELETONIZE_KWARGS}"
    )


def _extract_teasar_kwargs(kwargs: dict) -> dict:
    """Extract the teasar only kwargs"""
    return {key: value for key, value in kwargs.items() if key in TEASAR_KWARGS}



# Module for curved planar reformation from centerline. Only good for straightened CPR currently

import numpy as np
from scipy.interpolate import CubicSpline, interpn
from skimage import measure
from skimage.measure._regionprops import RegionProperties
from skimage.filters import gaussian
from skimage.morphology import remove_small_holes


###### CPR Generation Functions ###########
def create_straightened_cpr(
    vessel: np.ndarray,
    centerline: np.ndarray,
    cross_section_xy: tuple,
    resolution: int = 1,
    sigma: int = 2,
    is_binary: bool = True,
) -> np.ndarray:
    """Create a straightened Curved Planar Reformation of a vessel using a centerline
    :param vessel: the vessel as a numpy array
    :param centerline: the centerline as a numpy array
    :param cross_section_xy: the number of points in (x, y) for the cross-section;
        i.e. (100, 100) will result in an extracted cross-section of shape (100, 100)
    :param resolution: the resolution to sample the cpr. Default = 1
    :param sigma: the sigma value for the gaussian filter. Default = 2
    :return: a numpy array containing the straightened curved planar reformation
    """
    tangents = _compute_tangent_vectors(centerline)
    cpr = []
    for center_point, normal_vector in zip(centerline, tangents):
        cross_section = _extract_cross_sectional_slice(
            vessel, center_point, normal_vector, cross_section_xy, resolution, is_binary
        )
        if is_binary:
            cross_section = _postprocess_cross_section(cross_section, sigma)
        cpr.append(cross_section)
    cpr = np.stack(cpr, axis=0)
    return cpr


def _compute_tangent_vectors(centerline: np.ndarray) -> np.ndarray:
    """Compute the tangent vectors for a given centerline using a CubicSpline. Assumes that vertices are in natural order
    :param centerline: the centerline as an (N, 3) array
    :return: the tangent vectors for each vertex in the centerline
    """
    # Find the spline points
    s = np.zeros(len(centerline))
    s[1:] = np.cumsum(np.sqrt(np.sum(np.diff(centerline, axis=0) ** 2, axis=1)))
    # Generate the spline
    cs = CubicSpline(s, centerline, bc_type="natural")
    # Compute and normalize the tangents
    tangents = cs(s, 1)
    tangents /= np.linalg.norm(tangents, axis=1)[:, None]
    return tangents


def _compute_orthogonal_vectors(vector: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    """Find two orthogonal vectors to the original vector.
    :param vector: the vector to find the orthogonal vectors
    :return: the orthogonal vectors
    """
    if vector[0] == 0 and vector[1] == 0:
        if vector[2] == 0:
            # vec is a zero vector
            return None, None
        # vec is along the z-axis
        return np.array([1, 0, 0]), np.array([0, 1, 0])
    else:
        v1 = np.array([-vector[1], vector[0], 0])
        v1 = v1 / np.linalg.norm(v1)
        v2 = np.cross(vector, v1)
        return v1, v2


def _extract_cross_sectional_slice(
    arr, point, tangent_vector, slice_size, resolution, is_binary
):
    """
    Extract a 2D cross-sectional slice from a 3D array given a point and its tangent vector.

    :param arr: 3D binary NumPy array
    :param point: The point in the array (x, y, z)
    :param tangent_vector: The tangent vector at the point
    :param slice_size: The size of the slice to extract (width, height)
    :param resolution: The resolution of the slice (spacing between points in the grid)
    :param is_binary: Is the array binary or not
    """
    # Find two orthogonal vectors
    v1, v2 = _compute_orthogonal_vectors(tangent_vector)
    if v1 is None or v2 is None:
        raise ValueError("Invalid tangent vector provided.")
    # Generate a meshgrid for the slice
    width, height = slice_size
    x_lin = np.linspace(-width / 2, width / 2, int(width / resolution))
    y_lin = np.linspace(-height / 2, height / 2, int(height / resolution))
    x_grid, y_grid = np.meshgrid(x_lin, y_lin)
    # Map the grid points back to the 3D array indices
    slice_points = point + x_grid[..., np.newaxis] * v1 + y_grid[..., np.newaxis] * v2
    if is_binary:
        slice_points = np.rint(slice_points).astype(int)
        # Initialize an empty slice with zeros (padding)
        slice_2d = np.zeros(
            (int(height / resolution), int(width / resolution)), dtype=arr.dtype
        )
        if arr.min() != 0:
            slice_2d = slice_2d + arr.min()
        # Compute valid index ranges considering the boundaries
        valid_x = (slice_points[..., 0] >= 0) & (slice_points[..., 0] < arr.shape[0])
        valid_y = (slice_points[..., 1] >= 0) & (slice_points[..., 1] < arr.shape[1])
        valid_z = (slice_points[..., 2] >= 0) & (slice_points[..., 2] < arr.shape[2])
        valid_indices = valid_x & valid_y & valid_z
        # Extract values for valid indices and assign to the slice, leave zeros elsewhere
        valid_points = slice_points[valid_indices]
        slice_2d[valid_indices] = arr[
            valid_points[:, 0], valid_points[:, 1], valid_points[:, 2]
        ]
    else:
        # Create a grid of the original array coordinates
        x = np.arange(arr.shape[0])
        y = np.arange(arr.shape[1])
        z = np.arange(arr.shape[2])
        # Perform trilinear interpolation
        slice_2d = interpn(
            (x, y, z),
            arr,
            slice_points,
            method="linear",
            bounds_error=False,
            fill_value=arr.min(),
        )
    return slice_2d


def _postprocess_cross_section(cross_section: np.ndarray, sigma: int) -> np.ndarray:
    """Fill small gaps within the cross-section and apply a gaussian filter to smooth the edges
    :param cross_section: the extracted cross-section from _extract_cross_sectional_slice
    :param sigma: the gaussian kernel sigma
    :return: the smoothed cross-section
    """
    cross_section_labels = [label for label in np.unique(cross_section) if label > 0]
    output_cross_section = np.zeros_like(cross_section, dtype=np.uint8)
    for label in cross_section_labels:
        center_label = _filter_label_regions(cross_section, label)
        center_label = remove_small_holes(center_label)
        center_label = gaussian(center_label, sigma=sigma).round(0)
        output_cross_section[center_label == 1] = (
            label  # Assign the output of the binary to the appropriate label
        )
    return output_cross_section


def _filter_label_regions(cross_section: np.ndarray, label: int) -> np.ndarray:
    """Filter the labels within a CPR cross-section to only return the label region closest to the center of the image
    :param cross_section: the 2d numpy array containing the entire cross-section
    :param label: the label to filter
    :return: a numpy array containing the filtered cross-section
    """
    tmp_cross_section = (
        cross_section.copy()
    )  # Create a temporary array, so we don't change the original cross-section
    tmp_cross_section[tmp_cross_section != label] = (
        0  # set anything that is not the desired label to 0
    )
    tmp_cross_section[tmp_cross_section == label] = (
        1  # set the label to 1 for easier replacement later
    )
    regions = _get_regions(tmp_cross_section)
    if len(regions) == 0:
        centered_label = np.zeros_like(cross_section)
    elif len(regions) > 1:
        centered_label = _closest_to_centroid(tmp_cross_section, regions)
    else:
        centered_label = tmp_cross_section
    return centered_label > 0


def _closest_to_centroid(
    cross_section: np.ndarray, regions: list[RegionProperties]
) -> np.ndarray:
    """Filter a cross-section label using skimage.measure.regionprops to the region closest to the center
    :param cross_section: the cross-section array
    :param regions: the list output from skimage.measure.regionprops
    :return: the filtered numpy array
    """
    center_of_plane = np.array(cross_section.shape) / 2.0
    centroids = [np.array(region.centroid) for region in regions]
    distance_per_region = np.asarray(
        [np.linalg.norm(centroid - center_of_plane) for centroid in centroids]
    )
    min_distance_region_idx = int(np.argmin(distance_per_region))
    center_region = regions[min_distance_region_idx]
    center_label = np.zeros_like(cross_section)
    center_label[center_region.coords[:, 0], center_region.coords[:, 1]] = 1
    return center_label


######### CPR Measurement Functions #############
def measure_largest_cpr_diameter(
    cpr: np.ndarray, pixel_spacing: tuple, diff_threshold: int = 5
) -> dict:
    """Find the largest average cpr diameter
    :param cpr: a binary array containing the entire curved planar reformation to be measured
    :param pixel_spacing: the x, y pixel spacing for each cross-section
    :param diff_threshold: the maximum difference allowed between major and minor diameters for a cross-section.
        If |major - minor| > diff_threshold, set the average diameter = minor diameter to be conservative. Default = 5
    :return: a dictionary with {'avg_diam', 'major_diam', 'minor_diam'} as keys,
        holding the data from the largest diameter cross-section
    """
    avg_diams = []
    major_diams = []
    minor_diams = []
    for cross_section in cpr:
        avg_diam, major_diam, minor_diam = measure_cross_sectional_diameter(
            cross_section, pixel_spacing, diff_threshold
        )
        avg_diams.append(avg_diam)
        major_diams.append(major_diam)
        minor_diams.append(minor_diam)
    if avg_diams:
        largest_idx = np.argmax(avg_diams)
        diam_dict = {
            "avg_diam": avg_diams[largest_idx],
            "major_diam": major_diams[largest_idx],
            "minor_diam": minor_diams[largest_idx],
        }
    else:
        diam_dict = {}
    return diam_dict


def measure_mid_cpr_diameter(cpr: np.ndarray, pixel_spacing: tuple) -> dict:
    """Measure the mid-region diameter of a curved planar reformation
    :param cpr: a binary array containing the entire curved planar reformation to be measured
    :param pixel_spacing: the x, y pixel spacing for each cross-section
    :return: a dictionary with the mid-region average diameter"""
    len_cpr = len(cpr)
    midslice = len_cpr // 2
    if len_cpr > 0:
        avg_diam, _, _ = measure_cross_sectional_diameter(
            cpr[midslice], pixel_spacing, diff_threshold=5
        )
    else:
        avg_diam = None
    return {"mid_diam": avg_diam}


def measure_cross_sectional_diameter(
    cross_section: np.ndarray, pixel_spacing: tuple, diff_threshold: int
) -> tuple[float, float, float]:
    """Measure the cross-sectional diameter from a straightened cpr slice
    :param cross_section: the binary straightened cpr slice as a numpy array
    :param pixel_spacing: the pixel spacing of the cpr slice
    :param diff_threshold: the maximum difference allowed between major and minor diameters.
        If |major - minor| > diff_threshold, set the average diameter = minor diameter to be conservative
    :return: a tuple containing the average, max, and min diameter of the cpr slice
    """
    regions = _get_regions(cross_section)
    if len(regions) == 0:
        return 0, 0, 0
    region = regions[0]
    major_endpoints, minor_endpoints = _get_cross_section_endpoints(region)
    major_units = list(np.multiply(major_endpoints, pixel_spacing))
    minor_units = list(np.multiply(minor_endpoints, pixel_spacing))
    major_diam = _endpoint_euclidean_distance(major_units)
    minor_diam = _endpoint_euclidean_distance(minor_units)
    if abs(major_diam - minor_diam) < diff_threshold:
        avg_diam = round((major_diam + minor_diam) / 2, 1)
    else:
        avg_diam = min(major_diam, minor_diam)
    return avg_diam, major_diam, minor_diam


def _get_regions(image: np.ndarray) -> list[RegionProperties]:
    """Get the regions of an image using skimage.measure.regionprops
    :param image: the input image
    :return: the regions of an image
    """
    labels = measure.label(image)
    regions = measure.regionprops(labels)
    return regions


def _get_cross_section_endpoints(
    region: RegionProperties,
) -> tuple[tuple[tuple, tuple], tuple[tuple, tuple]]:
    centroid = region.centroid
    orientation = region.orientation
    major_endpoints = _get_axis_endpoints(
        centroid, orientation, region.axis_major_length
    )
    minor_endpoints = _get_axis_endpoints(
        centroid, orientation, region.axis_minor_length
    )
    return major_endpoints, minor_endpoints


def _get_axis_endpoints(
    centroid: np.array, orientation: float, axis_length: float
) -> tuple[tuple, tuple]:
    """Calculate the endpoints of the axis of a cross-section region
    :param centroid: the region.centroid
    :param orientation: the region.orientation
    :param axis_length: region.axis_major_length or region.axis_minor_length
    :return: a tuple containing the endpoints
    """
    y0, x0 = centroid
    # calculate the endpoints of the major axis using the centroid
    x1 = x0 - np.sin(orientation) * 0.5 * axis_length
    x2 = x0 + np.sin(orientation) * 0.5 * axis_length
    y1 = y0 - np.cos(orientation) * 0.5 * axis_length
    y2 = y0 + np.cos(orientation) * 0.5 * axis_length
    return (y1, x1), (y2, x2)


def _endpoint_euclidean_distance(endpoints: list) -> float:
    """Calculate the Euclidean distance between endpoints
    :param endpoints: the endpoints (must be two of them)
    :return: the Euclidean distance
    """
    p0, p1 = [*endpoints]
    return round(np.sqrt(((p0 - p1) ** 2).sum()), 1)



import SimpleITK as sitk
import numpy as np

from loguru import logger
from functools import partial

from mircato_stats.configs.models import torch_model_configs
from mircato_stats.configs.logging import timer
from mircato_stats.statistics.nifti import NiftiMircato
from mircato_stats.statistics.utils import _filter_largest_components
from mircato_stats.statistics.centerline import create_centerline
from mircato_stats.statistics.cpr import (
    create_straightened_cpr,
    measure_largest_cpr_diameter,
    measure_mid_cpr_diameter,
    _get_regions,
    measure_cross_sectional_diameter,
)


AORTA_CROSS_SECTION_SPACING = (1, 1)
ROOT_LENGTH = 10


@timer
def calculate_aorta_stats(nifti: NiftiMircato, vert_midlines: dict) -> dict:
    """Calculate the statistics for the aorta
    Parameters
    ----------
    nifti : NiftiMircato
        The nifti file to calculate statistics for
    vert_midlines : dict
        The vertebral midlines
    Returns
    -------
    dict[str: float]
        The statistics for the aorta
    """
    aorta_stats = {}
    # These are all the possible vertebral regions of the aorta. Scans will have some subset of these
    region_vert_map = {
        "abdominal": ["S1", *[f"L{i}" for i in range(1, 6)]],
        "thoracic": [f"T{i}" for i in range(1, 13)],
        "descending": [f"T{i}" for i in range(5, 13)],
    }
    # Define the output map
    # output_map = torch_model_configs["total"]["output_map"]
    # Get the regions to measure
    thoracic, abdominal, descending = _check_aorta_regions(vert_midlines)
    if not any((thoracic, abdominal, descending)):
        logger.warning("No aorta regions to measure")
        return aorta_stats
    # Get stats for each region of the aorta
    img = nifti.original_ct
    try:
        seg = _filter_to_aorta_seg(nifti.total_seg)
        measure_aorta_region = partial(_measure_aorta_region, img, seg)
        find_region_endpoints = partial(
            _find_aortic_region_endpoints, vert_midlines=vert_midlines
        )
    except Exception as e:
        logger.error(f"Error filtering to aorta segmentation: {e}")
        return aorta_stats
    
    if abdominal:
        try:
            start, end = find_region_endpoints(region_vert_map["abdominal"])
            abd_stats = measure_aorta_region(
                start, end, is_thoracic=False, region="abd_aorta"
            )
            aorta_stats.update(abd_stats)
        except Exception as e:
            logger.error(f"Error measuring abdominal aorta: {e}")
    if thoracic:
        try:
            start, end = find_region_endpoints(region_vert_map["thoracic"])
            thor_stats = measure_aorta_region(start, end, is_thoracic=True)
            aorta_stats.update(thor_stats)
        except Exception as e:
            logger.error(f"Error measuring thoracic aorta: {e}")
    elif descending:
        try:
            start, end = find_region_endpoints(region_vert_map["descending"])
            desc_stats = measure_aorta_region(
                start, end, is_thoracic=False, region="desc_aorta"
            )
            aorta_stats.update(desc_stats)
        except Exception as e:
            logger.error(f"Error measuring descending aorta: {e}")
    return aorta_stats


def _measure_aorta_region(
    img: sitk.Image,
    seg: sitk.Image,
    start: int,
    end: int,
    is_thoracic: bool,
    region: str | None = None,
) -> dict[str, float]:
    """Measure the aorta region in the image
    Parameters
    ----------
    img : sitk.Image
        The original image
    seg : sitk.Image
        The segmentation of the image
    start : int
        The start of the region to measure
    end : int
        The end of the region to measure
    is_thoracic : bool
        If the region is thoracic
    region : str
        The region to measure
    Returns
    -------
    dict[str: float]
        The statistics for the aorta region
    """
    # Get the region of the image to measure
    region_stats = {}
    # region_img = _aorta_superior(sitk.GetArrayFromImage(img[:, :, start:end]))
    region_seg = _aorta_superior(sitk.GetArrayFromImage(seg[:, :, start:end]))
    # Get the aorta segmentation
    aorta_label = 1
    aorta_anisotropy = (1, 1, 1)
    cross_section_size = (100, 100)
    cross_section_spacing = (1, 1)
    # Measure the diameters through a CPR
    centerline = create_centerline(
        region_seg, aorta_anisotropy, aorta_label, is_thoracic=is_thoracic
    )
    if centerline is None:
        logger.warning("No centerline found")
        return region_stats
    # Create the CPR
    cpr = create_straightened_cpr(
        region_seg, centerline, cross_section_xy=cross_section_size
    )
    if is_thoracic:
        diam_data = _measure_thoracic_diameters(region_seg, cpr, centerline)
        region_stats.update(diam_data)
    else:
        cpr = cpr[1:]  # remove the first slice to ignore oblong starting slices
        diam_data = measure_largest_cpr_diameter(cpr, cross_section_spacing)
        mid_diam = measure_mid_cpr_diameter(cpr, cross_section_spacing)
        diam_data.update(mid_diam)
        diam_data = {f"{region}_{k}": v for k, v in diam_data.items()}
        region_stats.update(diam_data)
    return region_stats


def _aorta_superior(arr: np.ndarray) -> np.ndarray:
    """Transform an array generated from a sitk image to the so that the arch of the aorta is at the top
    Parameters
    ----------
    arr : np.array
        The sitk -> numpy array to transform
    Returns
    -------
    np.array
        The LAS oriented array
    """
    arr = arr.transpose(2, 1, 0)
    arr = np.flip(np.rot90(arr, axes=(0, 2)), axis=1)
    return arr


def _filter_to_aorta_seg(total_seg: sitk.Image) -> sitk.Image:
    """Transform the total segmentation to only include the aorta and arch defining segmentations
    as well as perform morphological operations to properly orient the aorta
    Parameters
    ----------
    total_seg : sitk.Image
        The total segmentation
    Returns
    -------
    sitk.Image
        The aorta segmentation
    """
    output_map = torch_model_configs["total"]["output_map"]
    labels = ["aorta", "brachiocephalic_trunk", "subclavian_artery_left"]
    label_incides = [output_map[label] for label in labels]
    label_map = {
        old_label: new_label
        for new_label, old_label in enumerate(label_incides, start=1)
    }
    # Get the aorta segmentation
    aorta = sitk.GetArrayFromImage(total_seg)
    mask = np.isin(aorta, label_incides)
    aorta[~mask] = 0
    aorta_mapped = np.zeros_like(aorta)
    for label, new_label in label_map.items():
        aorta_mapped[aorta == label] = new_label
    mapped_indices = [int(x) for x in np.unique(aorta_mapped)]
    # Convert back to sitk image
    aorta_mapped = sitk.GetImageFromArray(aorta_mapped)
    aorta_mapped.CopyInformation(total_seg)
    # Remove erroneous segmentations
    aorta_mapped = _filter_largest_components(aorta_mapped, mapped_indices)
    return aorta_mapped


def _check_aorta_regions(vert_midlines: dict) -> tuple[bool, bool, bool]:
    """Check for the viability to measure the different aortic regions in the image
    Parameters
    ----------
    vert_midlines : dict
        The vertebral midlines
    Returns
    -------
    tuple[bool, bool, bool]
        The existince of the thoracic, abdominal, and descending aorta in the CT image
    """
    # T4 and at least T8 need to be in the image for thoracic to be measured
    thoracic = vert_midlines.get("vertebrae_T8_midline", False) and vert_midlines.get(
        "vertebrae_T4_midline", False
    )  # L3 has to be in the image for abdominal to be measured
    abdominal = vert_midlines.get("vertebrae_L3_midline", False)
    # If at least the T12 and T9 are in the image, then we can measure the descending separate from the ascending
    descending = vert_midlines.get(
        "vertebrae_T12_midline", False
    ) and vert_midlines.get("vertebrae_T9_midline", False)
    return thoracic, abdominal, descending


def _find_aortic_region_endpoints(verts: list, vert_midlines: dict) -> tuple[int, int]:
    """Parse the midlines from the vert dict to get the start and endpoints of the section
    Parameters
    ----------
    verts : list
        The list of vertebrae to measure
    vert_midlines : dict
        The vertebral midlines
    Returns
    -------
    tuple[int, int]
        The start and end of the region to measure
    """
    midlines = [
        vert_midlines.get(f"vertebrae_{vert}_midline")
        for vert in verts
        if vert_midlines.get(f"vertebrae_{vert}_midline") is not None
    ]
    midlines = [m for m in midlines if m is not None]
    start = min(midlines)
    end = max(midlines)
    return start, end


def _measure_thoracic_diameters(
    thoracic_aorta_seg: np.ndarray, cpr: np.ndarray, centerline: np.ndarray
) -> dict[str, float]:
    """Specialty function to measure maximum diameter of the ascending aorta, aortic arch and descending aorta.
    :param thoracic_aorta_seg: the entire aortic segmentation
    :param cpr: the curved planar reformation of the aorta
    :param centerline: the centerline of the thoracic aorta
    """
    unique_cpr_labels = np.unique(cpr).tolist()
    # Need background(0), aorta(1), brachiocephalic trunk(2) and left subclavian artery(3) in cpr
    # in order to do the simple split
    all_needed_labels = [0, 1, 2, 3]
    split_using_cpr = unique_cpr_labels == all_needed_labels
    if split_using_cpr:
        asc_cpr, arch_cpr, desc_cpr = _split_thoracic_using_cpr(cpr, centerline)
        if asc_cpr is None:  # If splitting using the CPR fails, split using the seg
            asc_cpr, arch_cpr, desc_cpr = _split_thoracic_using_seg(
                thoracic_aorta_seg, cpr, centerline
            )
    else:
        asc_cpr, arch_cpr, desc_cpr = _split_thoracic_using_seg(
            thoracic_aorta_seg, cpr, centerline
        )
    diam_data = {}
    for cpr, prefix in zip(
        [asc_cpr, arch_cpr, desc_cpr], ["asc_aorta", "aortic_arch", "desc_aorta"]
    ):
        if cpr is not None:
            region_diams = measure_largest_cpr_diameter(
                cpr, AORTA_CROSS_SECTION_SPACING
            )
            region_diams.update(
                measure_mid_cpr_diameter(cpr, AORTA_CROSS_SECTION_SPACING)
            )
            diams = {f"{prefix}_{k}": v for k, v in region_diams.items()}
            prox_diam, _, _ = measure_cross_sectional_diameter(
                cpr[0], AORTA_CROSS_SECTION_SPACING, diff_threshold=5
            )
            diams[f"{prefix}_prox_diam"] = prox_diam
            diam_data.update(diams)
    return diam_data


def _split_thoracic_using_cpr(cpr: np.ndarray, centerline: np.ndarray) -> tuple:
    """Split the thoracic cpr into the ascending, arch, and descending regions using the brachiocephalic trunk
    and left subclavian artery
    :param cpr: the cpr numpy array
    :param centerline: the centerline used to create the cpr
    :return: the ascending, arch, and descending cprs
    """
    brach_start, subclavian_end, aorta_label = _define_aortic_arch_with_seg(cpr)
    if subclavian_end < brach_start:
        logger.warning(
            "Left Subclavian Artery found above Brachiocephalic Trunk. use centerline to split instead."
        )
        return None, None, None
    # Separate each cpr
    asc_cpr, asc_centerline = cpr[:brach_start].copy(), centerline[:brach_start].copy()
    asc_cpr = _remove_aortic_root(asc_cpr, asc_centerline, ROOT_LENGTH)
    arch_cpr = cpr[brach_start:subclavian_end].copy()
    desc_cpr = cpr[subclavian_end:].copy()
    # Set all cprs to be binary (0 = background, 1 = aorta)
    asc_cpr[asc_cpr != aorta_label] = 0
    arch_cpr[arch_cpr != aorta_label] = 0
    desc_cpr[desc_cpr != aorta_label] = 0
    return asc_cpr, arch_cpr, desc_cpr


def _define_aortic_arch_with_seg(cpr):
    aorta_label = 1
    brach_label = 2
    subclavian_label = 3
    brach_start = 0
    subclavian_end = 0
    cpr_length = len(cpr)
    # First find the first instance of the brachiocephalic trunk
    for slice_idx, cross_section in enumerate(cpr):
        if brach_label in cross_section:
            brach_start = slice_idx
            break
    # Now find the last instance of the subclavian artery
    for slice_idx, cross_section in enumerate(
        cpr[::-1]
    ):  # Go through in reverse so first appearance == last instance
        if subclavian_label in cross_section:
            subclavian_end = cpr_length - slice_idx
            break
    return brach_start, subclavian_end, aorta_label


def _split_thoracic_using_seg(
    thoracic_aorta_seg: np.ndarray, cpr: np.ndarray, centerline: np.ndarray
) -> tuple:
    """The original way we split the thoracic aorta into ascending, arch, and descending regions using the segmentation
    :param thoracic_aorta_seg: the entire aortic segmentation
    :param cpr: the cpr numpy array
    :param centerline: the centerline used to create the cpr
    """
    # Find the highest point of the aortic centerline
    arch_peak = round(float(np.argmin(centerline, axis=0)[0]))
    asc_cpr = cpr[:arch_peak]
    asc_centerline = centerline[:arch_peak]
    desc_cpr = cpr[arch_peak:]
    desc_centerline = centerline[arch_peak:]
    # find where the aorta splits into 2 distinct regions axially
    # This defines the minimum
    min_pixel_area = 100
    split = None
    for slice_idx, axial_image in enumerate(thoracic_aorta_seg):
        regions = _get_regions(axial_image)
        if len(regions) == 2:
            reg0 = regions[0]
            reg1 = regions[1]
            # If both sections of the aorta are sufficiently large,
            if reg0.area > min_pixel_area and reg1.area > min_pixel_area:
                split = slice_idx
                break
    if split is None:
        logger.debug("Aortic arch could not be defined. Thoracic measurement skipped.")
        return None, None, None
    # now remove the arch based on the split
    # First ascending
    asc = []
    new_asc_centerline = []
    max_asc_idx = 0
    for i in range(len(asc_cpr)):
        if asc_centerline[i][0] >= split:
            asc.append(asc_cpr[i])
            new_asc_centerline.append(asc_centerline[i])
            max_asc_idx = i
    if len(asc) > 0:
        asc = np.stack(asc, axis=0)
        new_asc_centerline = np.stack(new_asc_centerline, axis=0)
        asc_cpr = _remove_aortic_root(asc, new_asc_centerline, ROOT_LENGTH)
    else:
        asc_cpr = None
    # descending
    desc = []
    min_desc_idx = len(desc_cpr) - 1
    for i in range(len(desc_cpr)):
        if desc_centerline[i][0] >= split:
            desc.append(desc_cpr[i])
            if (
                i < min_desc_idx
            ):  # This should only happen once, the first time the centerline is below the split
                min_desc_idx = i
    if len(desc) > 0:
        desc_cpr = np.stack(desc, axis=0)
    else:
        desc_cpr = None
    if max_asc_idx < min_desc_idx:
        arch_cpr = cpr[max_asc_idx:min_desc_idx]
    else:
        arch_cpr = None
    return asc_cpr, arch_cpr, desc_cpr


def _remove_aortic_root(
    ascending_cpr: np.ndarray,
    ascending_centerline: np.ndarray,
    threshold: int,
    invert: bool = False,
) -> np.ndarray:
    """Remove the aortic root region from the ascending aorta cpr
    :param ascending_cpr: the cpr of the ascending aorta
    :param ascending_centerline: the centerline of the ascending aorta
    :param threshold: the distance in mm on the centerline to skip (aka. defined length of aortic root)
    :param invert: Return the aortic root region instead of the region without the aortic root
    """
    distances = np.sum((ascending_centerline - ascending_centerline[0]) ** 2, axis=1)
    aortic_root_cutoff = np.where(distances >= (threshold**2))[0][0]
    if invert:
        return ascending_cpr[:aortic_root_cutoff]
    else:
        return ascending_cpr[aortic_root_cutoff:]



"""Contrast detection based on the Comp2Comp model with our segmentations"""

import os
import pickle
import numpy as np
import scipy
import SimpleITK as sitk
from scipy import ndimage as ndi

from loguru import logger
from warnings import filterwarnings
from mircato_stats.statistics.nifti import NiftiMircato
from mircato_stats.configs.models import torch_model_configs
from mircato_stats.configs.logging import get_project_root, timer

filterwarnings("ignore")
CONTRAST_PREDICTION_DICT = {0: "non_contrast", 1: "arterial", 2: "venous", 3: "delayed"}
LABEL_NAMES = [
    "aorta",
    "inferior_vena_cava",
    "portal_and_splenic_vein",
    "heart",
    "kidney_left",
    "kidney_right",
    "adrenal_gland_right",
]
FEATURE_LIST = [
    "aorta",
    "inferior_vena_cava",
    "portal_and_splenic_vein",
    "kidney_left",
    "kidney_right",
    "kidney_left_hull",
    "kidney_right_hull",
    "aorta_portal_vein",
    "aorta_ivc",
]


@timer
def predict_contrast(nifti: NiftiMircato) -> dict[str : str | str : float]:
    """Predict contrast phase of an image using radiomics data
    Parameters
    ----------
    nifti : NiftiMircato
        The nifti file to predict contrast for
    Returns
    -------
    dict
        The predicted contrast phase and the probability
    """
    try:
        # Load in the images and convert the segmentation to an array
        img = nifti.original_ct
        seg = nifti.total_seg
        seg_arr = sitk.GetArrayFromImage(nifti.total_seg)
        # Set the label map

        label_map = torch_model_configs["total"]["output_map"]
        binary_data = {}
        # Create all the binary masks
        label_names = LABEL_NAMES
        for name in label_names:
            label = label_map[name]
            binary_arr = _binarize_array(seg_arr, label)
            if name == "portal_and_splenic_vein":
                struct = np.ones((1, 1, 1))
            else:
                struct = np.ones((3, 3, 3))
            eroded_arr = ndi.binary_erosion(binary_arr, structure=struct).astype(np.uint8)
            if name == "portal_and_splenic_vein":
                if eroded_arr.sum() < 500:
                    eroded_arr = binary_arr
            eroded_sitk = sitk.GetImageFromArray(eroded_arr)
            eroded_sitk.CopyInformation(seg)
            binary_data[name] = eroded_sitk
            if name == "kidney_left" or name == "kidney_right":
                if eroded_arr.sum() > 0:
                    try:
                        mask = eroded_arr.copy()
                        hull = _fill_hull(mask)
                        hull = hull * (mask == 0)
                        struct = np.ones((3, 3, 3))
                        hull = ndi.binary_erosion(hull, structure=struct).astype(np.uint8)
                        hull_sitk = sitk.GetImageFromArray(hull)
                        hull_sitk.CopyInformation(seg)
                        binary_data[f"{name}_hull"] = hull_sitk
                    except Exception as e:
                        logger.error(e)
                        hull = np.zeros_like(eroded_arr)
                        hull_sitk = sitk.GetImageFromArray(hull)
                        hull_sitk.CopyInformation(seg)
                        binary_data[f"{name}_hull"] = hull_sitk
                else:
                    hull = np.zeros_like(eroded_arr)
                    hull_sitk = sitk.GetImageFromArray(hull)
                    hull_sitk.CopyInformation(seg)
                    binary_data[f"{name}_hull"] = hull_sitk
        # Get the statistics for each label
        statistics = {
            label: _get_sitk_stats(binary_data[label], img) for label in binary_data
        }
        statistics["aorta_portal_vein"] = (
            statistics["aorta"][:3] - statistics["portal_and_splenic_vein"][:3]
        )
        statistics["aorta_ivc"] = (
            statistics["aorta"][:3] - statistics["inferior_vena_cava"][:3]
        )
        features = []
        feature_list = FEATURE_LIST
        contrast_prediction_dict = CONTRAST_PREDICTION_DICT
        for feature in feature_list:
            features.extend(list(statistics[feature]))
        model = _load_model()
        try:
            y_pred_proba = model.predict_proba([features])[0]
            y_pred = np.argmax(y_pred_proba)
            pred_val = contrast_prediction_dict.get(y_pred)
            pred_prob = round(float(y_pred_proba[y_pred]), 3)
            return {"contrast_pred": pred_val, "contrast_prob": pred_prob}
        except Exception as e:
            logger.error(f"Contrast prediction failed: {e}")
            return {"contrast_pred": None, "contrast_prob": None}
    except Exception as e:
        logger.error(f"Contrast prediction failed: {e}")
        return {"contrast_pred": None, "contrast_prob": None}


def _load_model():
    model_path = os.path.join(get_project_root(), "models/xgboost.pkl")
    with open(model_path, "rb") as f:
        model = pickle.load(f)
    return model


def _binarize_array(arr: np.ndarray, label: int) -> np.ndarray:
    """Binarize a numpy array using a specified label"""
    return (arr == label).astype(np.uint8)


def _fill_hull(image: np.ndarray) -> np.ndarray:
    "Create the convex hull for a kidney segmentation"
    points = np.transpose(np.where(image))
    hull = scipy.spatial.ConvexHull(points)
    deln = scipy.spatial.Delaunay(points[hull.vertices])
    idx = np.stack(np.indices(image.shape), axis=-1)
    out_idx = np.nonzero(deln.find_simplex(idx) + 1)
    out_img = np.zeros(image.shape)
    out_img[out_idx] = 1
    return out_img


def _get_sitk_stats(seg: sitk.Image, img: sitk.Image):
    stats_filter = sitk.LabelIntensityStatisticsImageFilter()
    stats_filter.Execute(seg, img)
    if 1 not in stats_filter.GetLabels():
        return np.repeat(np.nan, 6)
    max_val = stats_filter.GetMaximum(1)
    min_val = stats_filter.GetMinimum(1)
    mean_val = stats_filter.GetMean(1)
    median_val = stats_filter.GetMedian(1)
    std_val = stats_filter.GetStandardDeviation(1)
    variance_val = stats_filter.GetVariance(1)
    return np.array([max_val, min_val, mean_val, median_val, std_val, variance_val])



"""
This file contains the implementation of the NiftiMircato class, which represents a folder of NIfTI files output from MirCATo and keeps state and meta information.

The NiftiMircato class provides methods for setting up the metadata for the NIfTI file, checking if it is valid for specific tasks, loading the original CT and segmentation arrays, and loading statistics.

Classes:
- BodySegNotFoundError: Raised when the body segmentation file is not found.
- TotalSegNotFoundError: Raised when the total segmentation file is not found.
- TissuesSegNotFoundError: Raised when the tissues segmentation file is not found.
- NiftiMircato: A class to represent a folder of NIfTI files output from MirCATo and keep state information.

Functions:
- resample_nifti_sitk: Load and resample a NIfTI file using SimpleITK.
"""

import json
import SimpleITK as sitk
from pathlib import Path
from loguru import logger

from mircato_stats.configs.statistics import stats_output_keys, midline_keys


class BodySegNotFoundError(FileNotFoundError):
    """
    Raised when the body segmentation file is not found
    """

    pass


class TotalSegNotFoundError(FileNotFoundError):
    """
    Raised when the total segmentation file is not found
    """

    pass


class TissuesSegNotFoundError(FileNotFoundError):
    """
    Raised when the tissues segmentation file is not found
    """

    pass


class NiftiMircato:
    """
    A class to represent a folder of NIfTI files output from MirCATo and keep state information.
    """

    task_labels = ["total", "tissues", "body"]

    def __init__(self, path: str):
        self.path = Path(path)  # assert that self.path is a Path object
        assert self.path.exists(), FileNotFoundError(
            f"Nifti {self.path} does not exist"
        )

    def __str__(self):
        return str(self.path)

    def setup(self, task_list: list[str]):
        """Setup the metadata for the NIfTI file and check if it is valid for tasks in task_list
        Parameters
        ----------
        task_list: list[str]
            A list of tasks to check for
        """
        # Set up the other meta information
        self.folder = self.path.parent
        self.nifti_name = self.path.name.partition(".")[0]
        self.seg_folder = self.folder.absolute() / f"{self.nifti_name}_segs"
        self.header_file = self.folder.absolute() / "header_info.json"
        assert self.seg_folder.exists(), FileNotFoundError(
            f"Segmentation folder {self.seg_folder} does not exist"
        )
        # Set up the stat attributes
        self.output_file = self.seg_folder / f"{self.nifti_name}_stats.json"

        # Create a dictionary for segmentation files
        self.seg_files = {
            task: self.seg_folder / f"{self.nifti_name}_{task}.nii.gz"
            for task in self.task_labels
        }
        self._check_seg_files(task_list)
        self._load_nifti_arrays(task_list)
        self._check_and_load_header()
        if self.output_file.exists():
            self._load_stats()
        else:
            self.stats_exist = False
            self.vert_midlines = {}

    def _check_and_load_header(self) -> None:
        if self.header_file.exists():
            try:
                with self.header_file.open("r") as f:
                    self.header_data = json.load(f)
            except json.decoder.JSONDecodeError:
                logger.warning("Header file {self.header_file} is corrupt or empty.")
                self.header_data = {}
        else:
            logger.warning(f"Header file {self.header_file} does not exist.")
            self.header_data = {}

        self.header_data["nifti_path"] = str(self.path.absolute())
        self.header_data["nii_file_name"] = str(self.nifti_name)

    def _check_seg_files(self, task_list: list[str]) -> None:
        """
        Check if the appropriate segmentation files exist for the given tasks
        Parameters
        ----------
        task_list: list[str]
            A list of tasks to check for
        """
        needs_total = any(
            [task in ["total", "aorta", "contrast"] for task in task_list]
        )
        needs_tissues = "tissues" in task_list
        if needs_total:
            if not self.seg_files["total"].exists():
                raise TotalSegNotFoundError(
                    f'Total segmentation file {self.seg_files["total"]} does not exist'
                )
        if needs_tissues:
            if not self.seg_files["body"].exists():
                raise BodySegNotFoundError(
                    f'Body segmentation file {self.seg_files["body"]} does not exist'
                )
            if not self.seg_files["tissues"].exists():
                raise TissuesSegNotFoundError(
                    f'Tissues segmentation file {self.seg_files["tissues"]} does not exist'
                )

    def _load_nifti_arrays(self, task_list, resample_spacing=[1.0, 1.0, 1.0]) -> None:
        """Load in the original CT and segmentation arrays
        Parameters
        ----------
        task_list: list[str]
            The list of statistics tasks
        resample_spacing: list[float]
            The spacing for the resampling
        """
        needs_total = any(
            [task in ["total", "aorta", "contrast"] for task in task_list]
        )
        needs_tissues = "tissues" in task_list
        self.original_ct = resample_nifti_sitk(
            self.path, resample_spacing, is_label=False
        )
        if needs_total:
            self.total_seg = resample_nifti_sitk(
                self.seg_files["total"], resample_spacing, is_label=True
            )
        if needs_tissues:
            self.body_seg = resample_nifti_sitk(
                self.seg_files["body"], resample_spacing, is_label=True
            )
            self.tissues_seg = resample_nifti_sitk(
                self.seg_files["tissues"], resample_spacing, is_label=True
            )
        # Check if the sizes and spacings are consistent
        if (
            self.original_ct.GetSize() != self.total_seg.GetSize()
            or self.original_ct.GetSize() != self.body_seg.GetSize()
            or self.original_ct.GetSize() != self.tissues_seg.GetSize()
            or self.original_ct.GetSpacing() != self.total_seg.GetSpacing()
            or self.original_ct.GetSpacing() != self.body_seg.GetSpacing()
            or self.original_ct.GetSpacing() != self.tissues_seg.GetSpacing()
        ):
            raise ValueError("Sizes and spacings of NIfTI arrays are not consistent")

    def _load_stats(self) -> None:
        try:
            with self.output_file.open("r") as f:
                stats = json.load(f)
            self.stats = stats
            vert_midlines = {k: stats.get(k) for k in midline_keys}
            self.vert_midlines = {
                k: v for k, v in vert_midlines.items() if v is not None
            }
            self.stats_exist = True

        except json.decoder.JSONDecodeError:
            logger.warning(f"Stats file {self.output_file} is corrupt or empty.")
            self.stats = {}
            self.stats_exist = False
            self.vert_midlines = {}

    def write_stats_to_file(self, output_stats: dict, all_completed: bool) -> None:
        """Write the statistics to a JSON file
        Parameters
        ----------
        output_stats: dict
            The statistics to write
        all_completed: bool
            Whether all statistics have been completed
        """
        if self.stats_exist:
            existing_stats = self.stats
            existing_stats.update(output_stats)
            output_stats = existing_stats
        if all_completed:
            flag_file = str(self.output_file).replace("_stats.json", ".complete")
            Path(flag_file).touch()
            output_stats = {k: output_stats.get(k) for k in stats_output_keys}
        with self.output_file.open("w") as f:
            json.dump(output_stats, f, indent=4)


def resample_nifti_sitk(
    nifti_path: Path, new_spacing: list[float], is_label: bool
) -> sitk.Image:
    """Load and resample a NIfTI file using SimpleITK
    Parameters
    ----------
    nifti_path: Path
        The path to the NIfTI file
    new_spacing: list[float]
        The new spacing for the resampling
    is_label: bool
        Whether the NIfTI file is a label map

    Returns
    -------
    sitk.Image
        The resampled image
    """
    input_image = sitk.ReadImage(nifti_path)
    # Get the original spacing and size
    original_spacing = input_image.GetSpacing()
    original_size = input_image.GetSize()

    # Calculate the new size
    new_size = [
        int(round(original_size[0] * (original_spacing[0] / new_spacing[0]))),
        int(round(original_size[1] * (original_spacing[1] / new_spacing[1]))),
        int(round(original_size[2] * (original_spacing[2] / new_spacing[2]))),
    ]
    interpolator = sitk.sitkNearestNeighbor if is_label else sitk.sitkLinear
    # Create the resampling filter
    resample = sitk.ResampleImageFilter()
    resample.SetOutputSpacing(new_spacing)
    resample.SetSize(new_size)
    resample.SetOutputDirection(input_image.GetDirection())
    resample.SetOutputOrigin(input_image.GetOrigin())
    resample.SetTransform(sitk.Transform())
    resample.SetDefaultPixelValue(input_image.GetPixelIDValue())

    resample.SetInterpolator(interpolator)

    # Perform the resampling
    return resample.Execute(input_image)



from time import time
from loguru import logger
from functools import partial
from pathlib import Path
from multiprocessing import Pool
from tqdm import tqdm
from .dicom_folder import DicomFolder as DicomFolder


def convert_dicom_folders_to_nifti(
    dicom_folders: list,
    output_dir: Path,
    num_workers: int,
    only_ax: bool,
    no_mip: bool,
    verbose: bool,
):
    """Main function to be passed to mircato CLI to convert DICOM folders to NIfTI
    Parameters
    ----------
    dicom_folders : list
        List of paths to DICOM folders
    output_dir : Path
        Path to the output directory
    num_workers : int
        Number of workers to use for the conversion
    only_ax : bool
        If True, only axial slices will be converted
    verbose : bool
        If True, verbose output will be printed
    """
    converter = partial(
        _process_dicom_folder, output_dir=output_dir, only_ax=only_ax, no_mip=no_mip
    )
    logger.info(
        f"Converting {len(dicom_folders)} dicom folders to nifti with {num_workers} workers"
    )
    start_time = time()
    with Pool(num_workers) as pool:
        if verbose:
            dicom_iterator = pool.imap_unordered(converter, dicom_folders)
        else:
            dicom_iterator = tqdm(
                pool.imap_unordered(converter, dicom_folders),
                total=len(dicom_folders),
                dynamic_ncols=True,
                desc="Converting DICOM folders to NIfTI",
            )
        for _ in dicom_iterator:
            pass
    end_time = time()
    logger.info(
        f"Conversion of {len(dicom_folders)} dicom folders to nifti completed in {end_time - start_time:.2f}s"
    )

def _process_dicom_folder(
    dicom_folder: str, output_dir: Path, only_ax: bool, no_mip: bool
) -> None:
    """Helper function to process a single dicom folder
    Parameters
    ----------
    dicom_folder : str
        Path to the dicom folder
    output_dir : Path
        Path to the output directory
    only_ax : bool
        If True, only axial slices will be converted
    no_mip : bool
        If True, mip series will not be converted
    """
    dicom_folder = DicomFolder(dicom_folder)
    dicom_folder.convert_to_nifti(output_dir, only_ax, no_mip)



from mircato_stats.configs.models import torch_model_configs
from mircato_stats.configs.logging import timer
from mircato_stats.statistics.nifti import NiftiMircato
from mircato_stats.statistics.utils import _calculate_3d_volumes, _calculate_2d_areas


@timer
def calculate_body_and_tissues_stats(
    nifti: NiftiMircato, vert_midlines: dict
) -> dict[str:float]:
    """Calculate the body and tissue statistics
    Parameters
    ----------
    nifti : NiftiMircato
        The nifti file to calculate statistics for
    vert_midlines : dict
        The vertebral midlines
    Returns
    -------
    dict
        The statistics for the body and tissues
    """
    body_map = torch_model_configs["body"]["output_map"]
    tissues_map = torch_model_configs["tissues"]["output_map"]
    # Set the segmentations
    body = nifti.body_seg
    tissues = nifti.tissues_seg
    # Get the midlines for the vertebrae of interest
    verts_of_interest = ["L1", "L3", "L5"]
    vert_indices = {
        v: vert_midlines.get(f"vertebrae_{v}_midline")
        for v in verts_of_interest
        if vert_midlines.get(f"vertebrae_{v}_midline") is not None
    }
    # Get the stats for the relevant regions
    output_stats = {}
    is_abdominal = "L1" in vert_indices and "L5" in vert_indices
    output_stats["abdominal_scan"] = int(is_abdominal)
    # Get the volumes for the entire scan
    prefix = "total_"
    body_volumes = _calculate_3d_volumes(body, body_map, prefix)
    tissues_volumes = _calculate_3d_volumes(tissues, tissues_map, prefix)
    output_stats.update(body_volumes)
    output_stats.update(tissues_volumes)
    # get the abdominal region values
    if is_abdominal:
        prefix = "abdominal_"
        endpoints = (
            vert_midlines.get("vertebrae_L5_midline"),
            vert_midlines.get("vertebrae_L1_midline") + 1,
        )
        abd_body_volumes = _calculate_3d_volumes(body, body_map, prefix, endpoints)
        abd_tissues_volumes = _calculate_3d_volumes(
            tissues, tissues_map, prefix, endpoints
        )
        output_stats.update(abd_body_volumes)
        output_stats.update(abd_tissues_volumes)
    # Get the vertebrae specific values
    for vert, midline in vert_indices.items():
        prefix = f"{vert}_"
        body_vert_stats = _calculate_2d_areas(
            body, body_map, midline, prefix, get_perimeter=True
        )
        tissues_vert_stats = _calculate_2d_areas(
            tissues, tissues_map, midline, prefix, get_perimeter=False
        )
        output_stats.update(body_vert_stats)
        output_stats.update(tissues_vert_stats)
    return output_stats



import argparse
from pathlib import Path
from mircato_stats.configs.logging import configure_logging, get_project_root
from mircato_stats.dicom import convert_dicom_folders_to_nifti
from mircato_stats.statistics import main as calculate_nifti_stats
import shutil


def mircato_stats():
    """
    MirCATo Stats Only CLI tool
    """
    parser = argparse.ArgumentParser(description="Mircato Stats Only CLI tool")
    parser.add_argument(
        "-q", "--quiet", help="Decrease output verbosity", action="store_true"
    )
    subparsers = parser.add_subparsers(dest="command", required=True)

    # Create the 'convert' subparser
    convert_parser = subparsers.add_parser("convert", help="Convert DICOM files")
    convert_parser.add_argument("dicoms", help="Path to DICOM files", type=Path)
    convert_parser.add_argument("output_dir", help="Output directory", type=Path)
    convert_parser.add_argument(
        "-n", "--num-workers", help="Number of something", type=int, default=1
    )
    convert_parser.add_argument(
        "-ax",
        "--axial-only",
        help="Only convert axial dicom series",
        action="store_true",
    )
    convert_parser.add_argument(
        "-nm", "--no-mip", help="Do not convert likely mip series", action="store_true"
    )
    # Set up stats parser
    stats_parser = subparsers.add_parser(
        "stats", description="Calculate statistics for NIfTI files"
    )
    stats_parser.add_argument(
        "niftis",
        type=Path,
        help="NIfTI file or a text file containing paths to multiple NIfTI files to calculate statistics for",
    )
    stats_parser.add_argument(
        "-t",
        "--task-list",
        type=str,
        nargs="+",
        default=["total", "contrast", "aorta", "tissues"],
        help='List of statistics tasks to perform. Default = ["total", "contrast", "aorta", "tissues"]',
    )
    stats_parser.add_argument(
        "-n",
        "--num-workers",
        type=int,
        default=4,
        help="Number of workers to use for multi-process operations. Default=1",
    )
    stats_parser.add_argument(
        "-th",
        "--threads",
        type=int,
        default=4,
        help="Number of threads each worker will use. Default=4",
    )
    stats_parser.add_argument(
        "-mc",
        "--mark-complete",
        action="store_true",
        help="Mark the statistics as complete regardless of stats performed",
    )
    args = parser.parse_args()
    args.verbose = not args.quiet
    if args.command == "convert":
        if args.dicoms.is_dir():
            logfile = "./nifti_conversion_log.json"
            dicom_list = [str(args.dicoms)]
        else:
            logfile = f'{args.dicoms.with_suffix("")}_conversion_log.json'
            with args.dicoms.open() as f:
                dicom_list = f.read().splitlines()
        configure_logging(logfile, args.verbose)
        convert_dicom_folders_to_nifti(
            dicom_list,
            args.output_dir,
            args.num_workers,
            args.axial_only,
            args.no_mip,
            args.verbose,
        )
    elif args.command == "stats":
        # If the input to the niftis argument is just a singular nifti file, make it a list and log in the same dir
        if args.niftis.suffixes == [".nii", ".gz"] or args.niftis.suffix == ".nii":
            logfile = str(args.niftis).replace(".nii", "").replace(".gz", "")
            logfile = f"{logfile}_{args.command}_log.jsonl"
            nifti_list = [str(args.niftis)]
        # Otherwise, make the log in the same dir as the input file and read the list
        else:
            logfile = f"{args.niftis.with_suffix('')}_{args.command}_log.jsonl"
            with args.niftis.open() as f:
                nifti_list = [x for x in f.read().splitlines()]
        configure_logging(logfile, args.verbose)

        calculate_nifti_stats(
                nifti_list,
                args.task_list,
                args.num_workers,
                args.threads,
                args.mark_complete,
            )
    else:
        print("Unknown command")


def copy_models():
    parser = argparse.ArgumentParser(description="Copy models to the correct location")
    parser.add_argument(
        "model_dir", help="Directory containing models to copy", type=Path
    )
    args = parser.parse_args()
    model_dir = args.model_dir
    if not model_dir.is_dir():
        raise ValueError("Model directory does not exist")
    project_root = get_project_root()
    destination_dir = project_root / "models"
    
    if not destination_dir.exists():
        destination_dir.mkdir(parents=True)
    
    for model_file in model_dir.iterdir():
        if model_file.is_file():
            shutil.copy(model_file, destination_dir)
    print(f"Models copied to {destination_dir}")



import sys
import json

from time import time
from pathlib import Path

from loguru import logger


# Quick timing wrapper
def timer(func):
    """Time a functions runtime
    :param func: the function to time
    :return: the time measured, and the result of the function
    """

    def wrap_func(*args, **kwargs):
        t1 = time()
        result = func(*args, **kwargs)
        t2 = time()
        # logger.info(f'{func.__name__!r} completed in {(t2-t1):.2f}s')
        return result, round(t2 - t1, 2)

    return wrap_func


# Logging functions
def get_project_root() -> Path:
    return Path(__file__).parent.parent


def serialize(record):
    subset = {
        "level": record["level"].name,
        "timestamp": record["time"].strftime("%Y-%m-%d %H:%M:%S"),
        "message": record["message"],
        # "exception": str(record["exception"]['type']),
        "module": record["module"],
        "line": record["line"],
        **record["extra"],
    }
    return json.dumps(subset)


def formatter(record):
    record["extra"]["serialized"] = serialize(record)
    return "{extra[serialized]}\n"


def configure_logging(log_file_path, verbose: bool):
    # Remove all handlers
    logger.remove()
    # Log all logs except INFO to a file
    logger.add(
        log_file_path,
        format=formatter,
        # serialize=True,
        level="DEBUG",
        rotation="10 GB",
        compression="gz",
        enqueue=True,
        retention=10,
        filter=lambda record: record["level"].name != "INFO",
    )
    if verbose:
        # Print INFO and SUCCESS logs to stdout
        logger.add(
            sys.stdout,
            colorize=True,
            format="<green>{time: DD-MM-YYYY -> HH:mm:ss}</green> <level>{message}</level>",
            level="INFO",
            filter=lambda record: record["level"].name in ["INFO", "SUCCESS"],
            enqueue=True,
        )
    else:
        # Only print INFO logs to stdout
        logger.add(
            sys.stdout,
            colorize=True,
            format="<green>{time: DD-MM-YYYY -> HH:mm:ss}</green> <level>{message}</level>",
            level="INFO",
            filter=lambda record: record["level"].name == "INFO",
            enqueue=True,
        )
    # Always print WARNING, ERROR, and CRITICAL to stderr
    logger.add(
        sys.stderr,
        colorize=True,
        format="<red>{time: DD-MM-YYYY -> HH:mm:ss}</red> <level>{message}</level>",
        level="WARNING",
        filter=lambda record: record["level"].name not in ["DEBUG", "INFO", "SUCCESS"],
        enqueue=True,
    )



diam_names = ["prox_diam", "mid_diam", "avg_diam", "major_diam", "minor_diam"]
stats_output_keys = [
    "nifti_path",
    "mrn",
    "accession",
    "series_name",
    "nii_file_name",
    "total_completed",
    "contrast_completed",
    "aorta_completed",
    "tissues_completed",
    "ct_direction",
    "image_type",
    "sex",
    "height_at_scan_m",
    "weight_at_scan_kg",
    "pregnancy_status",
    "birthday",
    "scan_date",
    "age_at_scan",
    "length_mm",
    "width_mm",
    "slice_thickness_mm",
    "num_rows",
    "num_columns",
    "manufacturer",
    "manufacturer_model",
    "kvp",
    "sequence_name",
    "protocol_name",
    "contrast_pred",
    "contrast_prob",
    "contrast_bolus_agent",
    "contrast_bolus_route",
    "multienergy_ct",
    "procedure_code",
    "procedure_desc",
    "study_uid",
    "series_uid",
    "is_mip",
    "spleen_volume_cm3",
    "spleen_average_intensity",
    "kidney_right_volume_cm3",
    "kidney_right_average_intensity",
    "kidney_left_volume_cm3",
    "kidney_left_average_intensity",
    "gallbladder_volume_cm3",
    "gallbladder_average_intensity",
    "liver_volume_cm3",
    "liver_average_intensity",
    "stomach_volume_cm3",
    "stomach_average_intensity",
    "pancreas_volume_cm3",
    "pancreas_average_intensity",
    "adrenal_gland_right_volume_cm3",
    "adrenal_gland_right_average_intensity",
    "adrenal_gland_left_volume_cm3",
    "adrenal_gland_left_average_intensity",
    "lung_upper_lobe_left_volume_cm3",
    "lung_upper_lobe_left_average_intensity",
    "lung_lower_lobe_left_volume_cm3",
    "lung_lower_lobe_left_average_intensity",
    "lung_upper_lobe_right_volume_cm3",
    "lung_upper_lobe_right_average_intensity",
    "lung_middle_lobe_right_volume_cm3",
    "lung_middle_lobe_right_average_intensity",
    "lung_lower_lobe_right_volume_cm3",
    "lung_lower_lobe_right_average_intensity",
    "esophagus_volume_cm3",
    "esophagus_average_intensity",
    "trachea_volume_cm3",
    "trachea_average_intensity",
    "thyroid_gland_volume_cm3",
    "thyroid_gland_average_intensity",
    "small_bowel_volume_cm3",
    "small_bowel_average_intensity",
    "duodenum_volume_cm3",
    "duodenum_average_intensity",
    "colon_volume_cm3",
    "colon_average_intensity",
    "bladder_volume_cm3",
    "bladder_average_intensity",
    "prostate_volume_cm3",
    "prostate_average_intensity",
    "sacrum_volume_cm3",
    "sacrum_average_intensity",
    "vertebrae_S1_volume_cm3",
    "vertebrae_S1_average_intensity",
    "vertebrae_L5_volume_cm3",
    "vertebrae_L5_average_intensity",
    "vertebrae_L4_volume_cm3",
    "vertebrae_L4_average_intensity",
    "vertebrae_L3_volume_cm3",
    "vertebrae_L3_average_intensity",
    "vertebrae_L2_volume_cm3",
    "vertebrae_L2_average_intensity",
    "vertebrae_L1_volume_cm3",
    "vertebrae_L1_average_intensity",
    "vertebrae_T12_volume_cm3",
    "vertebrae_T12_average_intensity",
    "vertebrae_T11_volume_cm3",
    "vertebrae_T11_average_intensity",
    "vertebrae_T10_volume_cm3",
    "vertebrae_T10_average_intensity",
    "vertebrae_T9_volume_cm3",
    "vertebrae_T9_average_intensity",
    "vertebrae_T8_volume_cm3",
    "vertebrae_T8_average_intensity",
    "vertebrae_T7_volume_cm3",
    "vertebrae_T7_average_intensity",
    "vertebrae_T6_volume_cm3",
    "vertebrae_T6_average_intensity",
    "vertebrae_T5_volume_cm3",
    "vertebrae_T5_average_intensity",
    "vertebrae_T4_volume_cm3",
    "vertebrae_T4_average_intensity",
    "vertebrae_T3_volume_cm3",
    "vertebrae_T3_average_intensity",
    "vertebrae_T2_volume_cm3",
    "vertebrae_T2_average_intensity",
    "vertebrae_T1_volume_cm3",
    "vertebrae_T1_average_intensity",
    "vertebrae_C7_volume_cm3",
    "vertebrae_C7_average_intensity",
    "vertebrae_C6_volume_cm3",
    "vertebrae_C6_average_intensity",
    "vertebrae_C5_volume_cm3",
    "vertebrae_C5_average_intensity",
    "vertebrae_C4_volume_cm3",
    "vertebrae_C4_average_intensity",
    "vertebrae_C3_volume_cm3",
    "vertebrae_C3_average_intensity",
    "vertebrae_C2_volume_cm3",
    "vertebrae_C2_average_intensity",
    "vertebrae_C1_volume_cm3",
    "vertebrae_C1_average_intensity",
    "heart_volume_cm3",
    "heart_average_intensity",
    "aorta_volume_cm3",
    "aorta_average_intensity",
    "pulmonary_vein_volume_cm3",
    "pulmonary_vein_average_intensity",
    "brachiocephalic_trunk_volume_cm3",
    "brachiocephalic_trunk_average_intensity",
    "subclavian_artery_right_volume_cm3",
    "subclavian_artery_right_average_intensity",
    "subclavian_artery_left_volume_cm3",
    "subclavian_artery_left_average_intensity",
    "carotid_artery_right_volume_cm3",
    "carotid_artery_right_average_intensity",
    "carotid_artery_left_volume_cm3",
    "carotid_artery_left_average_intensity",
    "brachiocephalic_vein_left_volume_cm3",
    "brachiocephalic_vein_left_average_intensity",
    "brachiocephalic_vein_right_volume_cm3",
    "brachiocephalic_vein_right_average_intensity",
    "atrial_appendage_left_volume_cm3",
    "atrial_appendage_left_average_intensity",
    "superior_vena_cava_volume_cm3",
    "superior_vena_cava_average_intensity",
    "inferior_vena_cava_volume_cm3",
    "inferior_vena_cava_average_intensity",
    "portal_and_splenic_vein_volume_cm3",
    "portal_and_splenic_vein_average_intensity",
    "iliac_artery_left_volume_cm3",
    "iliac_artery_left_average_intensity",
    "iliac_artery_right_volume_cm3",
    "iliac_artery_right_average_intensity",
    "iliac_vena_left_volume_cm3",
    "iliac_vena_left_average_intensity",
    "iliac_vena_right_volume_cm3",
    "iliac_vena_right_average_intensity",
    "vertebrae_S1_midline",
    *[f"vertebrae_L{i}_midline" for i in range(5, 0, -1)],
    *[f"vertebrae_T{i}_midline" for i in range(12, 0, -1)],
    *[f"vertebrae_C{i}_midline" for i in range(7, 0, -1)],
    *[f"asc_aorta_{diam}" for diam in diam_names],
    *[f"aortic_arch_{diam}" for diam in diam_names],
    *[f"desc_aorta_{diam}" for diam in diam_names],
    *[f"abd_aorta_{diam}" for diam in diam_names],
    # 'asc_root_fat_region_volume_cm3',
    # 'asc_root_fat_volume_cm3',
    # 'asc_root_mean_hu',
    # 'asc_root_stddev_hu',
    "abdominal_scan",
    "total_body_volume_cm3",
    "total_subq_fat_volume_cm3",
    "total_visc_fat_volume_cm3",
    "total_skel_muscle_volume_cm3",
    "abdominal_body_volume_cm3",
    "abdominal_subq_fat_volume_cm3",
    "abdominal_visc_fat_volume_cm3",
    "abdominal_skel_muscle_volume_cm3",
    "L1_body_border_ratio",
    "L1_body_area_cm2",
    "L1_body_total_perimeter_cm",
    "L1_body_ellipse_perimeter_cm",
    "L1_body_circle_perimeter_cm",
    "L1_subq_fat_area_cm2",
    "L1_visc_fat_area_cm2",
    "L1_skel_muscle_area_cm2",
    "L3_body_border_ratio",
    "L3_body_area_cm2",
    "L3_body_total_perimeter_cm",
    "L3_body_ellipse_perimeter_cm",
    "L3_body_circle_perimeter_cm",
    "L3_subq_fat_area_cm2",
    "L3_visc_fat_area_cm2",
    "L3_skel_muscle_area_cm2",
    "L5_body_border_ratio",
    "L5_body_area_cm2",
    "L5_body_total_perimeter_cm",
    "L5_body_ellipse_perimeter_cm",
    "L5_body_circle_perimeter_cm",
    "L5_subq_fat_area_cm2",
    "L5_visc_fat_area_cm2",
    "L5_skel_muscle_area_cm2",
]


midline_keys = [
    *[f"vertebrae_L{i}_midline" for i in range(5, 0, -1)],
    *[f"vertebrae_T{i}_midline" for i in range(12, 0, -1)],
    *[f"vertebrae_C{i}_midline" for i in range(7, 0, -1)],
]



from time import time
from functools import partial
from concurrent.futures import ThreadPoolExecutor
from numpy import array_split
from loguru import logger

from .nifti import NiftiMircato as NiftiMircato
from .aorta import calculate_aorta_stats as calculate_aorta_stats
from .total import calculate_total_segmentation_stats as calculate_total_stats
from .body_and_tissues import (
    calculate_body_and_tissues_stats as calculate_body_and_tissues_stats,
)
from .contrast_detection import predict_contrast as predict_contrast

from mircato_stats.configs import set_num_threads


def main(
    nifti_files: list[str],
    task_list: list[str],
    num_workers: int,
    num_threads: int,
    mark_complete: bool,
):
    """Function that calculates statistics for a list of nifti files over a set number of workers
    Parameters
    ----------
    nifti_files : list[str]
        The list of nifti files to process
    task_list : list[str]
        The list of statistics tasks to calculate
    num_workers : int
        The number of workers to use
    num_threads : int
        The number of threads to use
    mark_complete : bool
        Whether to mark the statistics as complete
    """
    logger.info(f'Processing statistics for {len(nifti_files)} nifti files')
    set_num_threads(num_threads)
    stats_start = time()
    if num_workers > 1:
        with ThreadPoolExecutor(max_workers=num_workers) as executor:
            futures = [executor.submit(single_nifti_stats, nifti, task_list, mark_complete) for nifti in nifti_files]
            for future in futures:
                future.result()
    else:
        for nifti in nifti_files:
            single_nifti_stats(nifti, task_list, mark_complete)
    stats_end = time()
    logger.info("Statistics processing completed in {:.2f}s".format(stats_end - stats_start))
    


def single_nifti_stats(
    input_nifti: str, task_list: list[str], mark_complete: bool
) -> None:
    """Calculate statistics for a single nifti file
    Parameters
    ----------
    nifti : Path
        The nifti file to calculate statistics for
    task_list : list[str]
        The list of statistics tasks to calculate
    mark_complete : bool
        Whether to mark the statistics as complete
    """
    nifti = NiftiMircato(input_nifti)
    nifti.setup(task_list)

    # Set up the default values
    header_data = nifti.header_data
    all_stats: dict = header_data
    contrast_time = 0
    total_time = 0
    aorta_time = 0
    tissues_time = 0
    total_stats = {}
    vert_midlines = {}
    all_completed = False
    # Run the statistics tasks
    if "contrast" in task_list:
        contrast_stats, contrast_time = predict_contrast(nifti)
        all_stats["contrast_completed"] = True
        all_stats.update(contrast_stats)

    if "total" in task_list:
        (total_stats, vert_midlines), total_time = calculate_total_stats(nifti)
        all_stats["total_completed"] = True
        all_stats.update(total_stats)
        all_stats.update(vert_midlines)

    if "aorta" in task_list:
        if "total" in task_list and total_stats and vert_midlines:
            if (
                total_stats.get("aorta_volume_cm3") is not None
                and all_stats.get("ct_direction", "AX") == "AX"
            ):
                aorta_stats, aorta_time = calculate_aorta_stats(nifti, vert_midlines)
                all_stats["aorta_completed"] = True
                all_stats.update(aorta_stats)
            else:
                logger.warning(
                    "Either aorta does not exist or the CT direction is not AX, so aorta will be skipped."
                )
        elif "aorta" in task_list and "total" not in task_list:
            if nifti.stats_exist and nifti.vert_midlines != {}:
                aorta_stats, aorta_time = calculate_aorta_stats(
                    nifti, nifti.vert_midlines
                )
                all_stats["aorta_completed"] = True
                all_stats.update(aorta_stats)
            else:
                logger.warning(
                    "Vertebrae midlines do not exist, so total stats will be ran."
                )
                (total_stats, vert_midlines), total_time = calculate_total_stats(nifti)
                all_stats["total_completed"] = True
                all_stats.update(total_stats)
                all_stats.update(vert_midlines)
                # Now run aorta stats
                if (
                    total_stats.get("aorta_volume_cm3") is not None
                    and all_stats.get("ct_direction", "AX") == "AX"
                ):
                    aorta_stats, aorta_time = calculate_aorta_stats(
                        nifti, vert_midlines
                    )
                    all_stats["aorta_completed"] = True
                    all_stats.update(aorta_stats)
                else:
                    logger.warning(
                        "Either aorta does not exist or the CT direction is not AX, so aorta will be skipped."
                    )

    if "tissues" in task_list:
        if vert_midlines:
            tissues_data, tissues_time = calculate_body_and_tissues_stats(
                nifti, vert_midlines
            )
            all_stats["tissues_completed"] = True
            all_stats.update(tissues_data)
        elif nifti.stats_exist and nifti.vert_midlines != {}:
            tissues_data, tissues_time = calculate_body_and_tissues_stats(
                nifti, nifti.vert_midlines
            )
            all_stats["tissues_completed"] = True
            all_stats.update(tissues_data)
        else:
            logger.warning(
                "Vertebrae midlines do not exist, so total stats will be ran."
            )
            (total_stats, vert_midlines), total_time = calculate_total_stats(nifti)
            all_stats["total_completed"] = True
            all_stats.update(total_stats)
            all_stats.update(vert_midlines)
            # Now run tissues stats
            tissues_data, tissues_time = calculate_body_and_tissues_stats(
                nifti, vert_midlines
            )
            all_stats["tissues_completed"] = True
            all_stats.update(tissues_data)

    if mark_complete:
        all_completed = True
        all_stats["total_completed"] = True
        all_stats["contrast_completed"] = True
        all_stats["aorta_completed"] = True
        all_stats["tissues_completed"] = True
    elif set(task_list) == set(["total", "contrast", "aorta", "tissues"]):
        all_completed = True
    elif nifti.stats_exist:
        all_stats.update(nifti.stats)
        if (
            all_stats.get("total_completed")
            and all_stats.get("contrast_completed")
            and all_stats.get("aorta_completed")
            and all_stats.get("tissues_completed")
        ):
            all_completed = True
    nifti.write_stats_to_file(all_stats, all_completed)
    logger.success(
        f"Stats for {nifti}, complete: {all_completed}. \n\tContrast Pred Time: {contrast_time}s. Total Stats Time: {total_time}s. Aorta Stats Time: {aorta_time}s. Tissue Stats time: {tissues_time}s.\n\tStats file -> {nifti.output_file}\n",
        extra={
            "key": "statistics",
            "input_nifti": str(nifti),
            "completed": True,
            "failed_reason": None,
            "output_file": str(nifti.output_file),
            "contrast_pred_time": contrast_time,
            "total_stats_time": total_time,
            "aorta_stats_time": aorta_time,
            "body_and_tissues_time": tissues_time,
        },
    )
